/*  DeltaASP: Delta-debugger for answer set solver development.
 *  Copyright (C) 2009  Robert Daniel Brummayer
 *
 *  This file is part of DeltaASP.
 *
 *  DeltaASP is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DeltaASP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package deltaasp;

import parser.*;
import java.io.*;
import java.util.*;

public class DeltaASP {

  String inName;

  String outName;

  String tmpName;
  
  Runtime rt;

  String []cmdArray;

  int goldenExitCode;

  int verbosity;

  long timeoutMillis;

  List<Rule> rules;

  boolean obo;

  int calls;

  static String strSuccess;

  static String strFail;

  static String strTimeout;

  static final int offset = 65;


  public DeltaASP(String inName, String outName, String tmpName, 
                  String []cmdArray, int verbosity, boolean colors,
                  long timeoutMillis, boolean obo){
    assert (inName != null);
    assert (outName != null);
    assert (tmpName != null);
    assert (cmdArray != null);
    assert (verbosity >= 0);
    assert (timeoutMillis >= 1000l);

    this.inName = inName;
    this.outName = outName;
    this.tmpName = tmpName;
    this.rt = Runtime.getRuntime();
    this.cmdArray = new String[cmdArray.length + 1];
    for (int i = 0; i < cmdArray.length; i++)
      this.cmdArray[i] = cmdArray[i];
    this.cmdArray[this.cmdArray.length - 1] = tmpName;
    this.verbosity = verbosity;
    this.timeoutMillis = timeoutMillis;

    if (colors) {
      this.strSuccess = (char) 27 + "[1;34m[" + (char) 27 + "[0;32mSUCCESS" +
                        (char) 27 + "[1;34m]" + (char) 27 + "[0;39m";
      this.strFail = (char) 27 + "[1;34m[" + (char) 27 + "[0;31mFAILED" + 
                     (char) 27 + "[1;34m]" + (char) 27 + "[0;39m";
      this.strTimeout = (char) 27 + "[1;34m[" + (char) 27 + "[1;30mTIMEOUT" +
                        (char) 27 + "[1;34m]" + (char) 27 + "[0;39m";
    } else {
      this.strSuccess = "[SUCCESS]";
      this.strFail = "[FAILED]";
      this.strTimeout = "[TIMEOUT]";
    }
    this.obo = obo;
    this.calls = 1; /* 1 because of call for golden exit code */
  }

  private static void copyFile (String dest, String src) throws IOException {
    BufferedReader in;
    PrintStream out;
    String s;

    assert (dest != null);
    assert (src != null);

    in = new BufferedReader (new FileReader(src));
    out = new PrintStream (dest);
    while ((s = in.readLine()) != null)
      out.println (s);
    in.close();
    out.close();
  }

  private void printlnVerbosityMsg (String msg, int verbosity) {
    if (this.verbosity >= verbosity) {
      System.out.println (msg);
      System.out.flush ();
    }
  }

  private void printVerbosityMsg (String msg, int verbosity) {
    if (this.verbosity >= verbosity)
      System.out.print (msg);
  }

  private void writeSimplifiedProgram ()
    throws IOException {
    PrintStream tmpStream;
    assert (this.rules != null);
    assert (this.rules.size() > 0);
    tmpStream = new PrintStream(this.tmpName);
    for (Rule rule : this.rules)
      tmpStream.println (rule.toString());
    tmpStream.close();
  }

  private SimplifiedResult runOnSimplifiedProgram () throws  IOException {
    SimplifiedResult result;
    int exitCode;
    Timer timer;
    Process proc;

    result = SimplifiedResult.NO;
    timer = new Timer();
    proc = this.rt.exec(this.cmdArray);
    timer.schedule (new Interrupter (Thread.currentThread()), 
                    this.timeoutMillis);
    try {
      exitCode = proc.waitFor();
      if (exitCode == this.goldenExitCode)
        result = SimplifiedResult.YES;
    } catch (InterruptedException ie) {
      result = SimplifiedResult.TIMEOUT;
    } finally {
      proc.destroy();
      timer.cancel();
    }
    return result;
  }

  private boolean checkSimplifiedProgram ()
    throws InterruptedException, IOException {
     
    boolean result = false;

    writeSimplifiedProgram ();
    switch (runOnSimplifiedProgram ()){
      case YES:
        printlnVerbosityMsg (strSuccess, 3);
        copyFile (this.outName, this.tmpName);
        result = true;
        break;
      case NO:
        printlnVerbosityMsg (strFail, 3);
        break;
      case TIMEOUT:
        printlnVerbosityMsg (strTimeout, 3);
        break;
    }

    this.calls++;
    return result;
  }

  private static long getFileSize (String fileName) throws IOException {
    File file;
    assert (fileName != null);
    file = new File (fileName);
    return file.length();
  }

  public void deltaDebug() throws ParserException, InterruptedException, 
    IOException {
    if (this.obo)
      this.deltaDebugOBO ();
    else
      this.deltaDebugDDMin();
  }

  private static <T> List<List<T>> splitList (List<T> list, int n){
    List<List<T>> result;
    int from, to,  numLiterals;
    List<T> sublist;

    assert (list != null);
    assert (n > 1);
    assert (list.size() > 0);

    numLiterals = list.size();
    result = new ArrayList<List<T>>(n);
    from = 0;
    for (int i = 0; i < n; i++) {
      to = from + ((numLiterals - from) / (n - i));
      sublist = list.subList(from, to);
      result.add (sublist);
      from += sublist.size();
    }
    return result;
  }

  private static <T> List<T> complementList (List<T> all, List<T> list){
    List<T> result;
    HashSet<T> set;

    assert (all != null);
    assert (list != null);
    assert (all.size() > 0);
    assert (list.size() > 0);
    assert (list.size() <= all.size());

    result = new ArrayList<T>();
    set = new HashSet<T>(list);
    for (T t : all)
      if (!set.contains(t))
        result.add (t);
    return result;
  }

  private <T> boolean ddMin(List<T> list, DDMinTest<T> ddminTest)
    throws InterruptedException, IOException {
    int n;
    int nextN = 0;
    boolean simplified = false;
    boolean foundNewSubset = false;
    boolean foundNewComplement = false;
    List<T> curList, curSublist, curSublistComp;
    List<List<T>> sublists;
    HashSet<List<T>> cache;

    n = 2;
    curList = list;
    cache = new HashSet<List<T>>();
    while (n <= curList.size()) {
      printlnVerbosityMsg("size of current set: " + curList.size(), 3);
      printlnVerbosityMsg("current granularity: " + n, 3);
      foundNewSubset = false;
      foundNewComplement = false;
      sublists = splitList (curList, n);
      assert (sublists != null);
      assert (sublists.size() > 0);

      for (int i = 0; i < sublists.size(); i++) {
        curSublist = sublists.get(i);
        if (cache.contains(curSublist))
          continue;
        cache.add(curSublist);
        if (this.verbosity >= 3)
          System.out.format ("%-" + offset + "s", "trying " + (i + 1) + 
                             ". subset of size " + 
                             curSublist.size() + ":");
        if (ddminTest.test(curSublist)) {
          nextN = 2;
          foundNewSubset = true;
          simplified = true;
          curList = curSublist;
          break;
        }
      }

      if (!foundNewSubset) {
        for (int i = 0; i < sublists.size(); i++) {
          curSublist = sublists.get(i);
          curSublistComp = complementList (curList, curSublist);
          if (cache.contains(curSublistComp))
            continue;
          cache.add(curSublistComp);
          if (this.verbosity >= 3)
            System.out.format ("%-" + offset + "s", "trying complement of " + 
                               (i + 1) + ". subset of size " + 
                               curSublistComp.size() + ":");
          if (ddminTest.test(curSublistComp)) {
            nextN = n - 1;
            foundNewComplement = true;
            simplified = true;
            curList = curSublistComp;
            break;
          }
        }
      }

      if (!foundNewSubset && !foundNewComplement) {
        if (n >= curList.size())
          break;
        nextN = n << 1;
        if (nextN > curList.size())
          nextN = curList.size();
      }

      n = nextN;
    }
    return simplified;
  }


  private void deltaDebugDDMin() throws ParserException, InterruptedException, 
    IOException {
    long origSize = 0;
    Process proc;
    List<RuleElement> body;
    List<Literal> lits;
    Literal lit;
    RuleElement head;
    RuleElement elem;
    boolean simplified = false;
    DisjunctiveHead dHead;
    CardinalityElement ce;
    WeightElement we;

    copyFile (tmpName, this.inName);
    copyFile (this.outName, this.inName);

    if (this.verbosity > 0) {
      printVerbosityMsg ("size of original program: ", 1);
      origSize = getFileSize (this.inName);
      printlnVerbosityMsg (origSize + " bytes", 1);
    }

    printVerbosityMsg ("computing golden exit code: ", 1);

    proc = rt.exec(this.cmdArray);
    this.goldenExitCode = proc.waitFor();
    printlnVerbosityMsg(String.valueOf(this.goldenExitCode), 1);

    printlnVerbosityMsg  ("parsing original program", 1);
    Parser.init(new BufferedReader (new FileReader(this.inName)));
    rules = Parser.parseASP();
    assert (rules != null);

    printlnVerbosityMsg  ("starting delta-debugging", 1);
    do {
      do {
        simplified = false;
        printlnVerbosityMsg  ("\nrunning ddmin on rule set", 2);
        simplified |= ddMin (this.rules, new DDMinTestRule (this, this.rules));

        printlnVerbosityMsg  ("\ntrying to delete bodies/heads", 2);
        for (Rule rule : this.rules) {
          body = rule.body;
          if (body != null && rule.head != null) {
            if (this.verbosity >= 3)
              System.out.format ("%-" + offset + "s", "trying to remove body:");
            rule.body = null;
            if (!checkSimplifiedProgram ()){
              rule.body = body;
              head = rule.head;
              rule.head = null;
              if (this.verbosity >= 3)
                System.out.format ("%-" + offset + "s", 
                                   "trying to remove head:");
              if (!checkSimplifiedProgram ())
                rule.head = head;
              else
                simplified = true;
            } else {
              simplified = true;
            }
          }
        }
      } while (simplified);

      printlnVerbosityMsg  ("\nrunning ddmin on individual literal sets", 2);

      for (Rule rule : this.rules) {
        head = rule.head;
        if (head != null) {
          if (head instanceof DisjunctiveHead) {
            printlnVerbosityMsg  ("trying to minimize disjunctive head:", 3);
            dHead = (DisjunctiveHead) head;
            simplified |= ddMin (dHead.lits, 
                                 new DDMinTestDHead (this, dHead.lits, dHead));
          } 
        } 
        body = rule.body;
        if (body != null) {
          printlnVerbosityMsg  ("trying to minimize body:", 3);
          simplified |= ddMin (body, new DDMinTestBody (this, body, rule));
        }
      }

      printlnVerbosityMsg  ("\nrunning ddmin on individual aggregates", 2);
      for (Rule rule : this.rules) {
        head = rule.head;
        if (head != null){
          if (head instanceof CardinalityElement) {
            printlnVerbosityMsg  ("trying to minimize cardinality " +
                                  "construct in head:", 3);
            ce = (CardinalityElement) head;
            simplified |= ddMin (ce.lits, 
                                 new DDMinTestCard (this, ce.lits, ce));
          } else if (head instanceof WeightElement) {
            printlnVerbosityMsg  ("trying to minimize weight " +
                                  "construct in head:", 3);
            we = (WeightElement) head;
            simplified |= ddMin (we.objs, 
                                 new DDMinTestWeight (this, we.objs, we));
          }
        }

        body = rule.body;
        if (body != null) {
          for (int j = body.size() - 1; j >= 0; j--) {
            elem = body.get(j);
            if (elem instanceof CardinalityElement) {
              printlnVerbosityMsg  ("trying to minimize cardinality " +
                                    "construct in body:", 3);
              ce = (CardinalityElement) elem;
              simplified |= ddMin (ce.lits, 
                                   new DDMinTestCard (this, ce.lits, ce));
            } else if (elem instanceof WeightElement) {
              printlnVerbosityMsg  ("trying to minimize weight " +
                                    "construct in body:", 3);
              we = (WeightElement) elem;
              simplified |= ddMin (we.objs, 
                                   new DDMinTestWeight (this, we.objs, we));
            }
          }
        }
      }

      printlnVerbosityMsg  ("\ntrying to remove negations", 2);
      for (Rule rule : this.rules) {
        head = rule.head;
        if (head != null){
          if (head instanceof WeightElement)
            simplified |= removeNegWeightElement ((WeightElement) head, 
                                                  this.verbosity);
          else if (head instanceof CardinalityElement)
            simplified |= removeNegCardElement ((CardinalityElement) head, 
                                                this.verbosity);
        }
        body = rule.body;
        if (body != null) {
          for (int j = body.size() - 1; j >= 0; j--) {
            elem = body.get(j);
            if (elem instanceof WeightElement)
              simplified |= removeNegWeightElement ((WeightElement) elem, 
                                                    this.verbosity);
            else if (elem instanceof CardinalityElement)
              simplified |= removeNegCardElement ((CardinalityElement) elem, 
                                                  this.verbosity);
            else if (elem instanceof Literal)
              simplified |= removeNegLit ((Literal) elem, this.verbosity);
          }
        }
      }
    } while (simplified);
    printStats (origSize);
  }

  private void printStats (long origSize) throws IOException  {
    long reducedSize;
    if (this.verbosity > 0) {
      printlnVerbosityMsg ("\noriginal program:   " + origSize + " bytes", 1);
      reducedSize = getFileSize (this.outName);
      printVerbosityMsg ("reduced program:    ", 1);
      System.out.format ("%" + String.valueOf (origSize).length() + 
                  "d bytes (",  reducedSize);
      System.out.format ("%.2f", ((double) reducedSize / origSize) * 100.0);
      printlnVerbosityMsg ("%)", 1);
      printlnVerbosityMsg ("executable calls:   " + this.calls, 1);
    }
  }

  private void deltaDebugOBO() throws ParserException, InterruptedException, 
    IOException {
    long origSize = 0;
    Process proc;
    List<RuleElement> body;
    List<Literal> lits;
    Literal lit;
    Rule rule;
    RuleElement head;
    RuleElement elem;
    int numRules;
    boolean simplified = false;

    copyFile (tmpName, this.inName);
    copyFile (this.outName, this.inName);

    if (this.verbosity > 0) {
      printVerbosityMsg ("size of original program: ", 1);
      origSize = getFileSize (this.inName);
      printlnVerbosityMsg (origSize + " bytes", 1);
    }

    printVerbosityMsg ("computing golden exit code: ", 1);

    proc = rt.exec(this.cmdArray);
    this.goldenExitCode = proc.waitFor();
    printlnVerbosityMsg(String.valueOf(this.goldenExitCode), 1);

    printlnVerbosityMsg  ("parsing original program", 1);
    Parser.init(new BufferedReader (new FileReader(this.inName)));
    this.rules = Parser.parseASP();
    assert (this.rules != null);
    numRules = this.rules.size();
    assert (numRules > 0);

    printlnVerbosityMsg  ("starting delta-debugging", 1);
    do {
      do {
        simplified = false;
        printlnVerbosityMsg  ("\ntrying to delete whole rules", 2);
        for (int i = numRules - 1; i >= 0 && this.rules.size() > 1; i--){
          rule = this.rules.remove (i);
          if (verbosity >= 3) {
            assert (!(rule.head == null && rule.body == null));
            if (rule.head == null)
              System.out.format ("%-" + offset + "s", 
                                 "trying to delete integrity constraint:");
            else if (rule.body == null)
              System.out.format ("%-" + offset + "s", "trying to delete fact:");
            else
              System.out.format ("%-" + offset + "s", 
                                 "trying to delete regular rule:");
          }
          if (!checkSimplifiedProgram ()) {
            this.rules.add(i, rule);
          } else {
            simplified = true;
          }
        }

        numRules = this.rules.size();
        printlnVerbosityMsg  ("\ntrying to delete bodies/heads", 2);
        for (int i = numRules - 1; i >= 0; i--) {
          rule = this.rules.get(i);
          body = rule.body;
          if (body != null && rule.head != null) {
            if (this.verbosity >= 3)
              System.out.format ("%-" + offset + "s", "trying to remove body:");
            rule.body = null;
            if (!checkSimplifiedProgram ()){
              rule.body = body;
              head = rule.head;
              rule.head = null;
              if (this.verbosity >= 3)
                System.out.format ("%-" + offset + "s", 
                                   "trying to remove head:");
              if (!checkSimplifiedProgram ())
                rule.head = head;
              else
                simplified = true;
            } else {
              simplified = true;
            }
          }
        }
      } while (simplified);

      printlnVerbosityMsg  ("\ntrying to delete individual literals", 2);
      for (int i = numRules - 1; i >= 0; i--) {
        rule = this.rules.get(i);
        head = rule.head;
        if (head != null){
          if (head instanceof DisjunctiveHead){
            lits = ((DisjunctiveHead) head).lits;
            for (int j = lits.size() - 1; j >= 0 && lits.size() > 1; j--) {
              if (this.verbosity >= 3)
                System.out.format ("%-" + offset + "s", 
                                   "trying to delete literal" +
                                   " from disjunctive head:");
              lit = lits.remove (j);
              if (!checkSimplifiedProgram ()) {
                lits.add(j, lit);
              } else {
                simplified = true;
              }
            }
          }
        }

        body = rule.body;
        if (body != null) {
          for (int j = body.size() - 1; j >= 0 && body.size() > 1; j--) {
            if (this.verbosity >= 3)
              System.out.format ("%-" + offset + "s", 
                                 "trying to delete literal" +
                                 " from body:");
            elem = body.remove(j);
            if (!checkSimplifiedProgram ()) {
              body.add(j, elem);
            } else {
              simplified = true;
            }
          }
        }
      }

      printlnVerbosityMsg  ("\ntrying to minimize aggregates", 2);
      for (int i = numRules - 1; i >= 0; i--) {
        rule = this.rules.get(i);
        head = rule.head;
        if (head != null){
          if (head instanceof WeightElement)
            simplified |= deleteWeightLits ((WeightElement) head, 
                                            this.verbosity);
          else if (head instanceof CardinalityElement)
            simplified |= deleteCardLits ((CardinalityElement) head, 
                                          this.verbosity);
        }

        body = rule.body;
        if (body != null) {
          for (int j = body.size() - 1; j >= 0; j--) {
            elem = body.get(j);
            if (elem instanceof WeightElement)
              simplified |= deleteWeightLits ((WeightElement) elem, 
                                              this.verbosity);
            else if (elem instanceof CardinalityElement)
              simplified |= deleteCardLits ((CardinalityElement) elem, 
                                            this.verbosity);
          }
        }
      }

      printlnVerbosityMsg  ("\ntrying to remove negations", 2);
      for (int i = numRules - 1; i >= 0; i--) {
        rule = this.rules.get(i);
        head = rule.head;
        if (head != null){
          if (head instanceof WeightElement)
            simplified |= removeNegWeightElement ((WeightElement) head, 
                                                  this.verbosity);
          else if (head instanceof CardinalityElement)
            simplified |= removeNegCardElement ((CardinalityElement) head, 
                                                this.verbosity);
        }
        body = rule.body;
        if (body != null) {
          for (int j = body.size() - 1; j >= 0; j--) {
            elem = body.get(j);
            if (elem instanceof WeightElement)
              simplified |= removeNegWeightElement ((WeightElement) elem, 
                                                    this.verbosity);
            else if (elem instanceof CardinalityElement)
              simplified |= removeNegCardElement ((CardinalityElement) elem, 
                                                  this.verbosity);
            else if (elem instanceof Literal)
              simplified |= removeNegLit ((Literal) elem, this.verbosity);
          }
        }
      }
    } while (simplified);
    printStats(origSize);
  }

  private boolean removeNegLit (Literal lit, int verbosity)
    throws InterruptedException, IOException {
    boolean simplified = false;

    assert (lit != null);
    assert (verbosity >= 0);

    if (lit.negAsFailure) {
      if (verbosity >= 3)
        System.out.format ("%-" + offset + "s", 
                           "trying to remove negation-as-failure");
      lit.negAsFailure = false;
      if (!checkSimplifiedProgram ())
        lit.negAsFailure = true;
      else
        simplified = true;
    }
    if (lit.negClassic) {
      if (verbosity >= 3)
        System.out.format ("%-" + offset + "s", 
                           "trying to remove classic negation");
      lit.negClassic = false;
      if (!checkSimplifiedProgram ())
        lit.negClassic = true;
      else
        simplified = true;
    }
    return simplified;
  }

  private boolean removeNegWeightElement (WeightElement elem, int verbosity)
    throws InterruptedException, IOException {
    List<WeightObject> objs;
    Literal lit;
    boolean simplified = false;

    assert (elem != null);
    assert (verbosity >= 0);

    objs = elem.objs;
    assert (objs != null);  
    for (int i = objs.size() - 1; i >= 0; i--) {
      lit = objs.get(i).lit;
      simplified |= removeNegLit (lit, verbosity);
    }
    return simplified;
  }

  private boolean removeNegCardElement (CardinalityElement elem, int verbosity)
    throws InterruptedException, IOException {
    List<Literal> lits;
    Literal lit;
    boolean simplified = false;

    assert (elem != null);
    assert (verbosity >= 0);

    lits = elem.lits;
    for (int i = lits.size() - 1; i >= 0; i--) {
      lit = lits.get(i);
      simplified |= removeNegLit (lit, verbosity);
    }
    return simplified;
  }

  private boolean deleteWeightLits (WeightElement elem, int verbosity)
    throws InterruptedException, IOException {
    List<WeightObject> objs;
    WeightObject obj;
    boolean simplified = false;

    assert (elem != null);
    assert (verbosity >= 0);

    objs = elem.objs;
    for (int i = objs.size() - 1; i >= 0 && objs.size() > 1; i--) {
      if (verbosity >= 3)
        System.out.format ("%-" + offset + "s", 
                           "trying to shrink weight construct");
      obj = objs.remove(i);
      if (!checkSimplifiedProgram ()) {
        objs.add (i, obj);
      } else {
        simplified = true;
      }
    }
    return simplified;
  }

  private boolean deleteCardLits (CardinalityElement elem, int verbosity)
    throws InterruptedException, IOException {
    List<Literal> lits;
    Literal lit;
    boolean simplified = false;

    assert (elem != null);
    assert (verbosity >= 0);

    lits = elem.lits;
    for (int i = lits.size() - 1; i >= 0 && lits.size() > 1; i--) {
      if (verbosity >= 3)
        System.out.format ("%-" + offset + "s", 
                           "trying to shrink cardinality construct");
      lit = lits.remove (i);
      if (!checkSimplifiedProgram ()) {
        lits.add(i, lit);
      } else {
        simplified = true;
      }
    }
    return simplified;
  }


  class Interrupter extends TimerTask {

    Thread t;

    Interrupter (Thread t) {
      assert (t != null);
      this.t = t;
    }

    public void run() { 
      t.interrupt(); 
    }

  }

  enum SimplifiedResult {
    YES, NO, TIMEOUT
  }


  abstract class DDMinTest<T> {

    DeltaASP deltaASP;

    List<T> list;

    DDMinTest () {
      this.deltaASP = null;
      this.list = null;
    }

    DDMinTest (DeltaASP deltaASP, List<T> list) {
      assert (deltaASP != null);
      assert (list != null);
      this.deltaASP = deltaASP;
      this.list = list;
    }
    
    abstract boolean test(List<T> l) throws InterruptedException, IOException;

  }

  class DDMinTestRule extends DDMinTest<Rule> {

    DDMinTestRule(DeltaASP deltaASP, List<Rule> list) {
      super (deltaASP, list);
    }

    boolean test(List<Rule> list) throws InterruptedException, IOException {
      List<Rule> prev;
      boolean result;

      assert (list != null);
      assert (list.size() > 0);

      prev = super.deltaASP.rules;
      super.deltaASP.rules = list;
      result = super.deltaASP.checkSimplifiedProgram(); 
      if (!result)
        this.deltaASP.rules = prev;
      return result;
    }
  }

  class DDMinTestDHead extends DDMinTest<Literal> {

    DisjunctiveHead head;

    DDMinTestDHead(DeltaASP deltaASP, List<Literal> list, 
                   DisjunctiveHead head) {
      super (deltaASP, list);
      assert (head != null);
      this.head = head;
    }

    boolean test(List<Literal> list) throws InterruptedException, IOException {
      List<Literal> prev;
      boolean result;

      assert (list != null);
      assert (list.size() > 0);

      prev = this.head.lits;
      this.head.lits = list;
      result = this.deltaASP.checkSimplifiedProgram(); 
      if (!result)
        this.head.lits = prev;
      return result;
    }
  }

  class DDMinTestBody extends DDMinTest<RuleElement> {

    Rule r;

    DDMinTestBody(DeltaASP deltaASP, List<RuleElement> list, Rule r) {
      super (deltaASP, list);
      assert (r != null);
      this.r = r;
    }

    boolean test(List<RuleElement> list) throws InterruptedException, 
                                                IOException {
      List<RuleElement> prev;
      boolean result;

      assert (list != null);
      assert (list.size() > 0);

      prev = this.r.body;
      this.r.body = list;
      result = this.deltaASP.checkSimplifiedProgram(); 
      if (!result)
        this.r.body = prev;
      return result;
    }
  }

  class DDMinTestCard extends DDMinTest<Literal> {

    CardinalityElement elem;

    DDMinTestCard(DeltaASP deltaASP, List<Literal> list, 
                   CardinalityElement elem) {
      super (deltaASP, list);
      assert (elem != null);
      this.elem = elem;
    }

    boolean test(List<Literal> list) throws InterruptedException, IOException {
      List<Literal> prev;
      boolean result;

      assert (list != null);
      assert (list.size() > 0);

      prev = this.elem.lits;
      this.elem.lits = list;
      result = this.deltaASP.checkSimplifiedProgram(); 
      if (!result)
        this.elem.lits = prev;
      return result;
    }
  }

  class DDMinTestWeight extends DDMinTest<WeightObject> {

    WeightElement elem;

    DDMinTestWeight(DeltaASP deltaASP, List<WeightObject> list, 
                    WeightElement elem) {
      super (deltaASP, list);
      assert (elem != null);
      this.elem = elem;
    }

    boolean test(List<WeightObject> list) 
      throws InterruptedException, IOException {
      List<WeightObject> prev;
      boolean result;

      assert (list != null);
      assert (list.size() > 0);

      prev = this.elem.objs;
      this.elem.objs = list;
      result = this.deltaASP.checkSimplifiedProgram(); 
      if (!result)
        this.elem.objs = prev;
      return result;
    }
  }


}
