/*  DeltaASP: Delta-debugger for answer set solver development.
 *  Copyright (C) 2009  Robert Daniel Brummayer
 *
 *  This file is part of DeltaASP.
 *
 *  DeltaASP is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DeltaASP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package parser;

import java.io.*;
import java.util.*;
import java.math.*;

public class Parser
{

  /* last recognized token */
  private static Token g_t;

  /* look ahead token */
  private static Token g_la;

  /* shortcut to kind attribute of lookahead token */
  private static TokenKind g_sym;

  private static List<Rule> g_rules;

  public static void init (BufferedReader in) {
    assert (in != null);
    reset ();
    Scanner.init(in);
    g_rules = new Vector<Rule>();
  }

  public static void reset (){
    Scanner.reset();
    g_t = null;
    g_la = null;
    g_sym = null;
    g_rules = null;
  }

  /* reads one symbol ahead */
  private static void scan() throws ParserException {
    g_t = g_la;
    try {
      g_la = Scanner.next();
    } catch (ScannerException se){
      throw new ParserException(se.getMessage());
    }
    g_sym = g_la.kind;
  }

  private static void parseError (String message, boolean onlyLine) 
    throws ParserException {
    String msg;
    assert (message != null);
    msg = g_la.line + ":";
    if (!onlyLine)
      msg += g_la.col;
    msg += " " + message;
    throw new ParserException (msg);
  }

  /* checks symbol and reads ahead */
  private static void check (TokenKind expected) throws ParserException {
    assert (expected != null);

    if (g_sym != expected && g_sym == TokenKind.EOF){
      throw new ParserException (g_la.line + ":" + g_la.col + 
                                 " unexpected end of file");
    }

    if (g_sym == expected)
      scan();
    else
      parseError ("expected: " + expected.getString(), false);
  }

  public static List<Rule> parseASP () throws ParserException {
    RuleElement head;
    List<RuleElement> body;
    Rule rule;

    scan();
    while (g_sym != TokenKind.EOF) {
      head = parseHead();
      if (g_sym == TokenKind.IF) {
        scan ();
        body = parseBody();
      } else {
        body = null;
      }
      check (TokenKind.DOT);
      g_rules.add (new Rule (head, body));
    }
    check (TokenKind.EOF);
    return g_rules;
  } 

  private static RuleElement parseHead () throws ParserException {
    RuleElement result = null;
    List<Literal> lits = new Vector<Literal>();
    boolean minus = false;
    switch (g_sym) {
      case IDENT: 
        lits.add(parseLiteral(false, true));
        while (g_sym == TokenKind.DIS) {
          scan();
          lits.add(parseLiteral (false, true));
        }
        if (lits.size() == 1)
          result = lits.get(0);
        else
          result = new DisjunctiveHead (lits);
        break;
      case MINUS:
        minus = true;
        scan ();
        if (g_sym != TokenKind.NUMERAL)
          parseError ("expected NUMERAL", false);
        /* fall through by intention */
      case NUMERAL:
      case LBRACE:
      case LBRACK:
        result = parseCardinalityWeightElement(false, minus);
        break;
      default:
        break;
    }
    return result;
  }

  private static RuleElement parseCardinalityWeightElement(boolean body, boolean minusLower)
    throws ParserException {
    RuleElement result;
    List<Boolean> weightIsSet = null;
    List<Integer> weights = null;
    List<Literal> lits = new Vector<Literal>();
    int upper = 0;
    int lower = 0;
    boolean upperIsSet = false;
    boolean lowerIsSet = false;
    boolean minus = false;
    boolean cardinality = false;
    int weight;
    assert (g_sym == TokenKind.MINUS || g_sym == TokenKind.LBRACE || 
            g_sym == TokenKind.NUMERAL || g_sym == TokenKind.LBRACK);

    if (g_sym == TokenKind.NUMERAL) {
      scan ();
      lower = Integer.valueOf(g_t.str).intValue();
      lowerIsSet = true;
      if (minusLower)
        lower = -lower;
    }

    if (g_sym != TokenKind.LBRACE && g_sym != TokenKind.LBRACK)
      parseError ("expected '{' or '['", false);
    if (g_sym == TokenKind.LBRACE) {
      cardinality = true;
      do {
        scan();
        lits.add(parseLiteral (body, true));
      } while (g_sym == TokenKind.COMMA);
      check (TokenKind.RBRACE);
    } else {
      assert (g_sym == TokenKind.LBRACK);
      weights = new Vector<Integer>();
      weightIsSet = new Vector<Boolean>();
      do {
        scan ();
        lits.add(parseLiteral (body, true));
        if (g_sym == TokenKind.EQ) {
          scan();
          minus = false;
          if (g_sym == TokenKind.MINUS) {
            scan();
            minus = true;
          }
          check (TokenKind.NUMERAL);
          weight = Integer.valueOf(g_t.str).intValue();
          if (minus)
            weight = -weight;
          weights.add (new Integer (weight));
          weightIsSet.add (Boolean.TRUE);
        } else {
          weights.add (new Integer(0));
          weightIsSet.add (Boolean.FALSE);
        }
      } while (g_sym == TokenKind.COMMA);
      check (TokenKind.RBRACK);
    }

    if (g_sym == TokenKind.MINUS) {
      if (cardinality)
        parseError ("upper bound of cardinality must not be < 0", false);
      scan ();
      check (TokenKind.NUMERAL);
      upper = Integer.valueOf(g_t.str).intValue();
      upper = -upper;
      upperIsSet = true;
    } else if (g_sym == TokenKind.NUMERAL) {
      scan ();
      upper = Integer.valueOf(g_t.str).intValue();
      upperIsSet = true;
    }
    if (body && !lowerIsSet && !upperIsSet)
      parseError ("need at least on bound for aggregate in body", true);

    if (lowerIsSet && upperIsSet && lower > upper)
      parseError ("lower bound must not be > upper bound", true);

    if (cardinality) {
      result =  new CardinalityElement (lits, lower, upper, lowerIsSet, 
                                        upperIsSet);
    } else  {
      result = new WeightElement (lits, weights, weightIsSet, lower, upper,
                                  lowerIsSet, upperIsSet); 
    }
    return result;
  }

  private static Literal parseLiteral (boolean negAsFailureAllowed,
                                       boolean negClassicAllowed) 
    throws ParserException {
    boolean negAsFailure = false;
    boolean negClassic = false;
    if (g_sym == TokenKind.NOT){
      if (!negAsFailureAllowed)
        parseError ("negation-as-failure not allowed here", false);
      negAsFailure = true;
      scan();
    }
    if (g_sym == TokenKind.MINUS){
      if (!negClassicAllowed)
        parseError ("classic negation not allowed here", false);
      negClassic = true;
      scan();
    }
    check (TokenKind.IDENT);
    return new Literal (g_t.str, negAsFailure, negClassic);
  }

  private static List<RuleElement> parseBody () throws ParserException {
    List<RuleElement> body = new Vector<RuleElement>();
    boolean negAsFailure = false;
    boolean negClassic = false;
    boolean minus = false;
    do {
      negAsFailure = false;
      negClassic = false;
      minus = false;
      switch (g_sym) {
        case MINUS:
          minus = true;
          scan ();
          if (g_sym != TokenKind.NUMERAL)
            parseError ("expected NUMERAL", false);
          /* fall through by intention */
        case NUMERAL:
        case LBRACE:
        case LBRACK:
          body.add(parseCardinalityWeightElement(true, minus));
          break;
        default:
          body.add (parseLiteral(true, true));
          break;
      }
      if (g_sym != TokenKind.COMMA)
        break;
      scan();
    } while (true);
    return body;
  }


}
