/*  DeltaASP: Delta-debugger for answer set solver development.
 *  Copyright (C) 2009  Robert Daniel Brummayer
 *
 *  This file is part of DeltaASP.
 *
 *  DeltaASP is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DeltaASP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package parser;

import java.io.*;
import java.util.*;

public class Scanner
{

  private static final char EOF ='\uffff'; /* used to represent EOF */

  private static final char CR = 13;

  private static final char LF = 10;

  private static Reader g_in;

  private static char g_ch;

  private static int g_line;

  private static int g_col;


  public static void init(Reader in){
    g_in = in;
    g_line = 1;
    g_col = 0;
    nextChar();
  }

  public static void reset() {
    g_in = null;
    g_ch = 0;
    g_line = 0;
    g_col = 0;
  }

  private static void nextChar (){
    int res;
    try {
      res = g_in.read();
      if (res < 0) {
        g_ch = EOF;
      } else {
        g_col++;
        g_ch = (char) res;
      }
      switch (g_ch){
        case CR:
          nextChar();
          break;
        case LF:
          g_line++;
          g_col = 0;
          break;
      }
    } catch (Exception e) {
      g_ch = EOF;
    }
  }


  private static String readIdentifierString () {
    StringBuilder builder = new StringBuilder();

    assert (Character.isLetter(g_ch) || g_ch == '_');

    do {
      builder.append(g_ch);
      nextChar();
    } while (Character.isLetterOrDigit(g_ch) || g_ch == '_');

    return builder.toString();
  }
  
  private static void readIdentifier (Token t) {
    TokenKind kind;
    String str;

    assert (Character.isLetter(g_ch));

    str = readIdentifierString();
    assert (str != null);
    assert (str.length() > 0);
    if (str.equals("not")) {
      t.kind = TokenKind.NOT;
    } else {
      t.kind = TokenKind.IDENT;
      t.str = str;
    }
  }

  private static void readNumeral (Token t) {
    StringBuilder builder = new StringBuilder();

    assert (Character.isDigit(g_ch));

    if (g_ch == '0') {
      builder.append(g_ch);
      nextChar();
    } else  {
      do {
        builder.append(g_ch);
        nextChar();
      } while (Character.isDigit(g_ch));
    }
    t.str = builder.toString();
    t.kind = TokenKind.NUMERAL;
  }

  public static Token next () throws ScannerException {
    while (Character.isWhitespace(g_ch))
      nextChar();

    Token result = new Token (TokenKind.NONE, g_line, g_col);
    switch (g_ch){
      case '(':
        result.kind = TokenKind.LPAR;
        nextChar();
        break;
      case ')':
        result.kind = TokenKind.RPAR;
        nextChar();
        break;
      case '[':
        result.kind = TokenKind.LBRACK;
        nextChar();
        break;
      case ']':
        result.kind = TokenKind.RBRACK;
        nextChar();
        break;
      case '{':
        result.kind = TokenKind.LBRACE;
        nextChar();
        break;
      case '}':
        result.kind = TokenKind.RBRACE;
        nextChar();
        break;
      case '.':
        result.kind = TokenKind.DOT;
        nextChar();
        break;
      case ',':
        result.kind = TokenKind.COMMA;
        nextChar();
        break;
      case '=':
        result.kind = TokenKind.EQ;
        nextChar();
        break;
      case '-':
        result.kind = TokenKind.MINUS;
        nextChar();
        break;
      case '|':
        result.kind = TokenKind.DIS;
        nextChar();
        break;
      case ':':
        nextChar();
        if (g_ch == '-') {
          result.kind = TokenKind.IF;
          nextChar();
        } else {
          result.kind = TokenKind.COLON;
        }
        break;
      case EOF:
        result.kind = TokenKind.EOF;
        break;
      default:
        if (Character.isLetter(g_ch)){
          readIdentifier (result);
        } else if (Character.isDigit(g_ch)){
          readNumeral(result);
        } else {
          throw new ScannerException (g_line + ":" + g_col +
                                      " unexpected character: \'" + 
                                      g_ch + "\'");
        }
        break;
    }

    return result;
  }

}
